#ask for the trusted security policy (hopefully, next release, the lhea policy
#will be an option, but it's currently broken)
policy trusted

#load the POW and fitsTcl features (they're built into the plugin as "static
#packages")
load "" pow
load "" fits

#make a place to put the POW window
frame .powholder 
grid .powholder -column 0 -row 0 -columnspan 2

#open up a POW window to plot stuff in
#arguments mean:
# "safe" - tell POW we're running in a browser so don't try to manage the X
#          color resources
# ".powholder"    - where to put the POW window
# "0"             - don't include the standard POW GUI (You might try 
#                   setting this to "1" for a more versatile tclet
powInit safe .powholder 1

#tell the POW window how big to be
.pow.pow configure -width 380 -height 380

button .imagebutton -text "Plot Image" -command {loadImage $imageURL}
grid .imagebutton -column 0 -row 1 -sticky ew

set imageURL "http://heasarc.gsfc.nasa.gov/wasabi/algol.fits"
entry .imageentry -textvariable imageURL -width 40
grid .imageentry -column 1 -row 1 -sticky ew


proc loadImage {url} {
    
#make a name for the graph:
set gname "MyGraph"
#make sure we don't reuse an existing name
set ext  0
while {[lsearch [powListGraphs] $gname] != -1} {
    incr ext
    set gname "${gname}_${ext}"
}


#for simplicity, in this example, we'll let cfitsio read the file.
#if you're running under Windows or MacOS, this means you can't use
#a URL, just a local file name. We could be more elaborate, but this
#is meant to be a *simple* example
#open the fits file (readonly)    
set infilehandle [fits open $url 0]

#load the image data into memory
set imghandle [$infilehandle load image]
    
#get the dimensions of the image
set dims [$infilehandle info imgdim]
set n1 [lindex $dims 0]
set n2 [lindex $dims 1]

#get the data type of the image 
set data_type [lindex [lindex [$infilehandle get keyword BITPIX] 0] 1]

#Now we work around a bug in the 2.0 version of the plugin, future versions
#will be able to use the BITPIX value for data type without translation
    set ppFitsDataType(8) 0
    set ppFitsDataType(16) 1
    set ppFitsDataType(32) 2
    set ppFitsDataType(-32) 3
    set ppFitsDataType(-64) 4
    set data_type $ppFitsDataType($data_type)


#Now a bit of Voodoo to deal with possible wierd file types:

#If the image has BZERO or BSCALE keywords in the header, fitsTcl will
#do the appropriate thing with them automatically, but the datatype returned
#will be floating point doubles (isn't FITS fun:)
if { ([catch {$infilehandle get keyword BZERO}] == 0) ||
     ([catch {$infilehandle get keyword BSCALE}] == 0) } {
    set data_type 4
}

#make a POW DATA object

powCreateData ${gname}_data $imghandle $data_type [expr $n1 * $n2] 0

#make a POW IMAGE object; the units (pixels, intensity) are arbitrary; since
#this is a general application, we don't know what they are

powCreateImage ${gname}_img ${gname}_data 0 0 $n1 $n2 0 1 0 1 pixels pixels intensity

#This will setup POW to use the Astronomical coordinate information
#in the file (if there is any) 
global powWCS
if { ! [catch  {$infilehandle get imgwcs} wcsString] } {
    set powWCS(${gname}_img) $wcsString
}

#we're done reading the file now
$infilehandle close

#make a POW Graph object which contains only our image

powCreateGraph $gname NULL ${gname}_img \
    pixels pixels NULL NULL 300 300

}
