set helpMsg "Usage: POWplot ?-cmap code? ?fitsimage?
 code: 0 Chose the best colormap
       1 Force to install private pseudo colormap
       2 Force to use truecolor colormap (if don't have it will crash!) 
       3 Force to use screen default colormap"


set ppFitsDataType(8) 0
set ppFitsDataType(16) 1
set ppFitsDataType(32) 2
set ppFitsDataType(-32) 3
set ppFitsDataType(-64) 4

set instPos [lsearch $argv "-cmap"]
set colorCode 0

# if install colormap
if { $instPos != -1} {
    if { [catch {set colorCode [lindex $argv [expr $instPos+1]]}] == 1} {
	puts $helpMsg
	exit
    }
    if { ($colorCode != 0) && ($colorCode != 1) && ($colorCode != 2) \
	&& ($colorCode != 3) } {
	puts $helpMsg
	exit
    }	
    set argv [lreplace $argv $instPos [expr $instPos+1]]
    incr argc -2 
}

wm withdraw .
update idletask

global powWCS
global POW_LIBRARY
set POW_LIBRARY "$env(POW_LIBRARY)/.."
# auto_load 
load [glob $POW_LIBRARY/libfitstcl.s\[ol\]*]
load [glob $POW_LIBRARY/libpow.s\[ol\]*]

# try to avoid the flash. 
powSetupColormap .dummy 30 $colorCode
wm withdraw .dummy

powInit .dummy

foreach filename $argv {
    regsub -all {[^a-zA-Z0-9]} $filename "_" gname
    set filehandle [fits open $filename 0]
    set imghandle [$filehandle load image]
    set dims [$filehandle info imgdim]
    set n1 [lindex $dims 0]
    if { [llength $dims]==1 } {
       set n2 1
    } else {
       set n2 [lindex $dims 1]
    }
    set data_type [lindex [lindex [$filehandle get keyword BITPIX] 0] 1]
    set data_type $ppFitsDataType($data_type)
    if { ([catch {$filehandle get keyword BZERO}] == 0) ||
         ([catch {$filehandle get keyword BSCALE}] == 0) } {
        set data_type 4
    }
    powCreateData ${gname}_data $imghandle $data_type [expr $n1 * $n2] 0

    # Try to get WCS (or scaling) information

    set x0 1
    set y0 1
    set xinc 1
    set yinc 1
    if { ! [catch  {$filehandle get imgwcs} wcsString] } {
       if { [lindex $wcsString 7] == "none" } {
	  foreach {xval yval xref yref xinc yinc} [lrange $wcsString 0 5] {}
	  # locate center of pixel 1,1
	  set x0 [expr $xval + (1.0-$xref)*$xinc]
	  set y0 [expr $yval + (1.0-$yref)*$yinc]
       } else {
	  set powWCS(${gname}_img) $wcsString
       }
    }

    powCreateImage ${gname}_img ${gname}_data \
	0 0 $n1 $n2 $x0 $xinc $y0 $yinc pixels pixels counts
    powCreateGraph $gname NULL ${gname}_img \
	pixels pixels NULL NULL 300 300
    $filehandle close
}


bind . <<powExit>> {exit}

